#ifndef LEVEL_NUM_COUNTER_H
#define LEVEL_NUM_COUNTER_H

#include <map>
#include <numeric>
#include <functional>
#include <iterator>


template <
    class Key,
    class T,
    class Compare = std::less<Key>,
    class Allocator = std::allocator<std::pair<const Key, T> >
> class LevelNumCounter
{
    
public:
    typedef typename std::map<Key, T>::value_type value_type;
    typedef typename std::map<Key, T>::size_type size_type;
    typedef typename std::map<Key, T>::iterator iterator;
    typedef typename std::map<Key, T>::const_iterator const_iterator;

public:
    
    LevelNumCounter() : count_m() { }
    
    /*!
     * Add more "particles" to that level
     * @param level where to add
     * @param nTimes to increment
     */
    void increment(const Key& level, T nTimes = T(1)) { count_m[level] += nTimes; }
    
    /*!
     * Add more "particles" to that level
     * @param level where to add
     * @param nTimes to decrement
     */
    void decrement(const Key& level, T nTimes = T(1)) { increment(level, -nTimes); }
    
    T& operator[](T level) { return count_m[level]; }
    
    const T& operator[](T level) const { return count_m[level]; }
    
    size_type size() const { return count_m.size(); }
    
    bool empty() const { return count_m.empty(); }
    
    iterator begin() { return count_m.begin(); }
    const_iterator begin() const { return count_m.begin(); }
    
    iterator end() { return count_m.end(); }
    const_iterator end() const { return count_m.end(); }
    
    
    /*!
     * Obtain the start of a level
     * @param level
     * @returns the local starting index
     */
    T begin(T level) const {
        auto end = count_m.begin();
        
        // make sure to stay within container
        T size = count_m.size();
        std::advance(end, (level > size) ? size : level);
        
        return std::accumulate(count_m.begin(), end, 0,
                               [](T sum, const value_type& value_pair) {
                                   return sum + value_pair.second;
                               });
    }
    
    
    /*!
     * Obtain the end of a level
     * @param level
     * @returns the index of the local end of that level
     */
    T end(T level) const { return begin(level + 1); }
    
    
    /*!
     * Remove particle indices from the container
     * @param num of particles that will be removed
     * @param begin of index
     */
    void remove(T num, T begin) {
        int inum = int(num);
        while ( inum > -1 ) {
            T level = which(begin + inum);
            --count_m[level];
            --inum;
        }
    }
    
    /*!
     * @returns the total count of particles over
     * all levels
     */ 
    T getLocalNumAllLevel() {
        return begin(count_m.size());
    }
    
private:
    /*!
     * Find the level the particle belongs to
     * @param idx is the local index of the particle
     * @returns the level
     */
    T which(T idx) {
        T level = 0;
        
        while ( idx >= end(level) && level < size() )
            ++level;
        
        return level;
    }
        
        
private:
    std::map<Key, T> count_m;
};

#endif